--[[-----------------------------------------------------------------------------
Support for LDB tooltips handled via OnEnter/OnLeave
-------------------------------------------------------------------------------]]

local addonName, addon = ...
setfenv(1, addon)

local GetMouseFocus, GetPoint, MouseIsOver = GetMouseFocus, UIParent.GetPoint, MouseIsOver

local MONITOR_INTERVAL = 0.05

--[[-----------------------------------------------------------------------------
Helpers
-------------------------------------------------------------------------------]]
local tooltipCache, activeTooltips, lastFrame = { }

local function FindTooltip(plugin, tooltips)
	local relFrame
	for _, tooltip in pairs(tooltips) do
		_, relFrame = GetPoint(tooltip, 1)
		if relFrame == plugin then
			return tooltip
		end
	end
end

local function UpdateTooltipCache()
	local EnumerateFrames, IsObjectType = EnumerateFrames, UIParent.IsObjectType
	local frame = EnumerateFrames(lastFrame)
	while frame do
		if IsObjectType(frame, 'GameTooltip') then
			tooltipCache[frame] = frame
		end
		frame, lastFrame = EnumerateFrames(frame), frame
	end
end

--[[-----------------------------------------------------------------------------
Monitor
-------------------------------------------------------------------------------]]
local monitor = CreateFrame('Frame', nil, UIParent)
monitor:Hide()

do
	local delay, timer, timeOut = 0, 0, 0

	local isValidFocus = setmetatable({ }, {
		__index = function(self, frame)
			if frame == WorldFrame then
				for index = 1, #self do
					if MouseIsOver(self[index]) then
						return true
					end
				end
			elseif frame then
				local _, relFrame = GetPoint(frame, 1)
				if relFrame and relFrame ~= WorldFrame then
					local isValid = self[relFrame] or false
					self[frame] = isValid
					return isValid
				end
			end
			return false
		end,

		__newindex = function(self, frame, isValid)
			if frame ~= WorldFrame then
				rawset(self, frame, isValid)
				if isValid and not frame:IsMouseEnabled() then
					local left, bottom, width, height = frame:GetRect()
					if left and bottom and width and height then
						local right, top = left + width, bottom + height
						for index = 1, #self do
							local L, B, W, H = self[index]:GetRect()
							if L and B and W and H and L >= left and B >= bottom and L + W <= right and B + H <= top then
								return
							end
						end
					end
					rawset(self, #self + 1, frame)
				end
			end
		end
	})

	monitor:SetScript('OnHide', function(self)
		local tooltip = self.tooltip
		if tooltip.SetAutoHideDelay then
			tooltip:SetAutoHideDelay()
		end
		if tooltip.Release then
			tooltip:Release()
		else
			tooltip:Hide()
		end
		self.plugin:DetachFrame()
		self.plugin, self.tooltip = nil, nil
		wipe(isValidFocus)
	end)

	monitor:SetScript('OnShow', function(self)
		timer, timeOut = 0, 0
		self.tooltip:EnableMouse(true)															-- Interactive tooltips should be doing this already but...
		isValidFocus[self.plugin] = true
		isValidFocus[self.tooltip] = true
		isValidFocus[UIParent] = false

		local autoHide = self.tooltip.autoHideTimerFrame
		if autoHide then
			delay = autoHide.delay
			if autoHide.alternateFrame and autoHide.alternateFrame ~= self.plugin then
				isValidFocus[autoHide.alternateFrame] = true
			end
			if autoHide.parent and autoHide.parent ~= self.tooltip then
				isValidFocus[autoHide.parent] = true
			end
		else
			delay = 0
		end
	end)

	monitor:SetScript('OnUpdate', function(self, elapsed)
		timer = timer + elapsed
		if timer < MONITOR_INTERVAL then return end

		if isValidFocus[GetMouseFocus()] then
			timeOut = 0
		else
			timeOut = timeOut + timer
			if timeOut >= delay then
				self:Hide()
			end
		end
		timer = 0
	end)
end

--[[-----------------------------------------------------------------------------
Global to addon
-------------------------------------------------------------------------------]]
function AcquireTooltip(plugin)
	local tooltip
	if activeTooltips then
		tooltip = FindTooltip(plugin, activeTooltips)
	end
	if not tooltip then
		UpdateTooltipCache()
		tooltip = FindTooltip(plugin, tooltipCache)
		if tooltip and tooltip:GetAnchorType() ~= 'ANCHOR_NONE' then
			tooltip:SetAnchorType('ANCHOR_NONE')
		end
	end
	if tooltip then
		monitor:Hide()
		plugin:AttachFrame(tooltip)
		monitor.plugin, monitor.tooltip = plugin, tooltip
		monitor:Show()
	end
end

function ReleaseTooltip(plugin)
	if plugin then
		local tooltip = plugin.attached
		if tooltip and tooltip:IsShown() then
			return
		end
	end
	monitor:Hide()
end

--[[-----------------------------------------------------------------------------
Initialize
-------------------------------------------------------------------------------]]
monitor:SetScript('OnEvent', function(self, event)
	local LQT = LibStub('LibQTip-1.0', true)
	if not LQT then return end
	activeTooltips = LQT.activeTooltips
	self:SetScript('OnEvent', nil)
	self:UnregisterEvent(event)
end)
monitor:RegisterEvent('ADDON_LOADED')

UpdateTooltipCache()
